/**************************************************************************
* iptrk - ip tracker
* Copyright (C) 2003 Jason Nunn <jsno@leal.com.au>
* Darwin, NT, Australia.
* Originally developed for Arafura Connect (Dec 2000 - Mar 2003)
* Donated to GPL before liquidation.
*
* =======================================================================
* data organisation routines
*
**************************************************************************/
#include "defs.h"
#include "lib.p"

PUBLIC treap_t *watch_bt,
               *our_net_ip_bt;
PUBLIC long    watch_bt_size;
PUBLIC int     watch_bt_size_lim_f;

PRIVATE FILE   *file_fp;

/**************************************************************************
* binary atom routines
**************************************************************************/
PUBLIC int cmp_16(U_16 d0, U_16 d1)
{
  S_32 x;

  x = (S_32)d0 - (S_32)d1;
  if(x < 0)
    return -1;
  if(x > 0)
    return 1;

  return 0;
}

PUBLIC int cmp_32(U_32 d0, U_32 d1)
{
  S_64 x;

  x = (S_64)d0 - (S_64)d1;
  if(x < 0)
    return -1;
  if(x > 0)
    return 1;

  return 0;
}

PUBLIC int cmp_64(U_64 d0, U_64 d1)
{
  S_64 x;

  x = (S_64)(d0 >> 32)        - (S_64)(d1 >> 32);           /*ms*/
  if(x < 0)
    return -1;
  if(x > 0)
    return 1;

  x = (S_64)(d0 & 0xffffffff) - (S_64)(d1 & 0xffffffff);    /*ls*/
  if(x < 0)
    return -1;
  if(x > 0)
    return 1;

  return 0;
}

PUBLIC int port_append_cmp_func(void *n1, void *n2)
{
  return cmp_32(((port_t *)n1)->port.key, ((port_t *)n2)->port.key);
}

PUBLIC int iptrk_append_cmp_func(void *n1, void *n2)
{
  return cmp_64(((iptrk_t *)n1)->addr.key, ((iptrk_t *)n2)->addr.key);
}

/**************************************************************************
* gather local IP's
**************************************************************************/
PRIVATE void update_addr(treap_t **bt, U_32 addr)
{
  U_32 *p;

  int find_cmp_func(void *n)
  {
    return cmp_32(addr, *((U_32 *)n));
  }

  int append_cmp_func(void *n1, void *n2)
  {
    return cmp_32(*((U_32 *)n1), *((U_32 *)n2));
  }

  p = treap_find(bt, find_cmp_func);
  if(p == NULL)
  {
    treap_append(bt, &addr, sizeof(U_32), append_cmp_func);

    p = treap_find(bt, find_cmp_func);
    if(p == NULL)
    {
      log_m('E', __func__, "::treap_find(): failed");
      exit(-1);
    }
  }
}

PUBLIC void gather_local_ips(treap_t *bt, treap_t **ip_tree)
{
  void s_tree(treap_t *t)
  {
    iptrk_t *n;

    if(t == NULL)
      return;

    s_tree(t->l);

    n = TREAP_ENTRY(t);
    if(get_our_net(n->addr.addr.s, NULL))
      update_addr(ip_tree, n->addr.addr.s);
    if(get_our_net(n->addr.addr.d, NULL))
      update_addr(ip_tree, n->addr.addr.d);

    s_tree(t->r);
  }

  s_tree(bt);
}

/**************************************************************************
*
**************************************************************************/
PUBLIC int get_our_net(U_32 a, our_net_t **p)
{
  ll_t      *ll;
  our_net_t *n;

  ll = boot.our_net_ll;
  for(;;)
  {
    n = ll_read_entry(&ll);
    if(n == NULL)
      break;

    if(n->addr.addr == (a & n->addr.mask))
    {
      if(p != NULL)
        *p = n;
      return 1;
    }
  }

  if(p != NULL)
    *p = NULL;
  return 0;
}

PUBLIC inline void voldr_acc(voldr_t *a, voldr_t *p)
{
  a->I += p->I;
  a->O += p->O;
  a->u += p->u;
  a->i += p->i;
  a->o += p->o;
}

/**************************************************************************
* cast volume
*
* this is for log casting
**************************************************************************/
PUBLIC void cast_volume_log_sub(
  U_32 src, U_32 dst, addr_t *addr, U_64 vol, voldr_t *voldr, int dir)
{
  memset(voldr, 0, sizeof(voldr_t));

  switch(dir)
  {
    case DIR_I:
      voldr->I = vol;
      break;

    case DIR_O:
      voldr->O = vol;
      break;

    case DIR_u: /*internal traffic*/
      if((dst & addr->mask) == addr->addr)
      {
        if((src & addr->mask) == addr->addr)
        {
          voldr->u = vol;
        }
        else
        {
          dir = DIR_i;
          voldr->i = vol;
        }
      }
      else
      {
        if((src & addr->mask) == addr->addr)
        {
          dir = DIR_o;
          voldr->o = vol;
        }
      }
      break;
  }
}

PUBLIC int cast_volume_log(
  U_32 src, U_32 dst, addr_t *addr, U_64 vol, voldr_t *voldr)
{
  int d;

  d = (get_our_net(src, NULL) << 1) | get_our_net(dst, NULL);
  cast_volume_log_sub(src, dst, addr, vol, voldr, d);

  return d;
}

/**************************************************************************
* cast volume
*
* this is a special cast the for monitoring sub system
**************************************************************************/
PUBLIC int cast_volume_monitor(U_32 src, U_32 dst, addr_t *addr)
{
  int dir;

  dir = (get_our_net(src, NULL) << 1) | get_our_net(dst, NULL);
  if(dir == DIR_u)
  {
    if((dst & addr->mask) == addr->addr)
    {
      if((src & addr->mask) != addr->addr)
      {
        dir = DIR_i;
      }
    }
    else
    {
      if((src & addr->mask) == addr->addr)
      {
        dir = DIR_o;
      }
    }
  }

  return dir;
}

/**************************************************************************
* tree (watch) update routines
**************************************************************************/
PRIVATE void update_port_entry(
  treap_t **t, U_16 port_s, U_16 port_d, U_16 p_vol)
{
  port_t entry, *p;

  int find_cmp_func(void *n)
  {
    return cmp_32(entry.port.key, ((port_t *)n)->port.key);
  }

  entry.port.port.s = port_s;
  entry.port.port.d = port_d;

  p = treap_find(t, find_cmp_func);
  if(p == NULL)
  {
    entry.vol = 0;
    treap_append(t, &entry, sizeof(port_t), port_append_cmp_func);

    p = treap_find(t, find_cmp_func);
    if(p == NULL)
    {
      log_m('E', __func__, "::treap_find(): failed");
      exit(-1);
    }
  }

  p->vol += p_vol;
}

PUBLIC void update_ip_entry(
  treap_t **bt,
  U_32 src, U_32 dst, U_16 vol, U_8 prot, U_16 port_s, U_16 port_d)
{
  iptrk_t entry, *p;

  int find_cmp_func(void *n)
  {
    return cmp_64(entry.addr.key, ((iptrk_t *)n)->addr.key);
  }

  entry.addr.addr.s = src;
  entry.addr.addr.d = dst;
  p = treap_find(bt, find_cmp_func);
  if(p == NULL)
  {
    entry.oth_port_ll = NULL;
    entry.tcp_port_ll = NULL;
    entry.udp_port_ll = NULL;
    entry.vol = 0;
    entry.created = timer;
    treap_append(bt, &entry, sizeof(iptrk_t), iptrk_append_cmp_func);
    watch_bt_size++;

    p = treap_find(bt, find_cmp_func);
    if(p == NULL)
    {
      log_m('E', __func__, "::treap_find(): failed");
      exit(-1);
    }
  }

  p->vol += vol;
  p->last_updated = timer;

  switch(prot)
  {
    case IPPROTO_TCP:
      update_port_entry(&(p->tcp_port_ll), port_s, port_d, vol);
      break;

    case IPPROTO_UDP:
      update_port_entry(&(p->udp_port_ll), port_s, port_d, vol);
      break;

    default:
      update_port_entry(&(p->oth_port_ll), 0, 0, vol);
      break;
  }
}

/**************************************************************************
* save routines
**************************************************************************/
PRIVATE void save_tree_port(treap_t *t)
{
  if(t != NULL)
  {
    port_t *e;

    save_tree_port(t->l);

    e = TREAP_ENTRY(t);
    fprintf(file_fp,"%hx %hx %llx\n", e->port.port.s, e->port.port.d, e->vol);

    save_tree_port(t->r);
  }
}

PRIVATE void save_tree_ip(treap_t *t)
{
  if(t != NULL)
  {
    iptrk_t *n;

    save_tree_ip(t->l);

    n = TREAP_ENTRY(t);

    fprintf(file_fp,"%lx %lx %llx %lx %lx\n",
      n->addr.addr.s, n->addr.addr.d, n->vol,
      n->created, n->last_updated);

    fprintf(file_fp, "O\n");
    save_tree_port(n->oth_port_ll);
    putc('\n', file_fp);

    fprintf(file_fp, "T\n");
    save_tree_port(n->tcp_port_ll);
    putc('\n', file_fp);

    fprintf(file_fp, "U\n");
    save_tree_port(n->udp_port_ll);
    putc('\n', file_fp);

    save_tree_ip(t->r);
  }
}

PUBLIC void save_file(treap_t *bt, char *filename)
{
  char str[STR_LEN + 1];

  sprintf(str, " saving db file %s", filename);
  log_m('I', NULL, str);

  file_fp = fopen(filename, "w");
  if(file_fp == NULL)
  {
    log_e(__func__, "::fopen()");
    exit(-1);
  }

  save_tree_ip(bt);
  fclose(file_fp);
}

/**************************************************************************
* aliasizes a port value
**************************************************************************/
PUBLIC U_16 mod_port(U_16 in)
{
  eserv_t *p;

  int cmp_func(void *n)
  {
    return cmp_16(in, ((eserv_t *)n)->port);
  }

  if(in < 1024)
    return in;

  p = treap_find(&(boot.eserv_bt), cmp_func);
  if(p == NULL)
    return 0;

  return p->port;
}

/**************************************************************************
*
**************************************************************************/
PUBLIC int find_debug_addr(U_32 src, U_32 dst)
{
  addr_key_t e;

  int cmp_func(void *n)
  {
    return cmp_64(e.key, ((addr_key_t *)n)->key);
  }

  e.addr.s = src;
  e.addr.d = dst;
  if(treap_find(&(boot.debug_bt), cmp_func) == NULL)
    return 0;

  return 1;
}

/**************************************************************************
*
**************************************************************************/
PUBLIC void save_treaps(void)
{
  char str[STR_LEN + 1], date_str[STR_LEN + 1];

  file_date(time_st, date_str);
  sprintf(str, "%s/db.%s", IPTRK_DB_WATCH_PATHC, date_str);
  save_file(watch_bt, str); 
}
